#include <avr\io.h>
#include <avr\interrupt.h>

#define F_CPU 16000000
#include <util/delay.h>
#include <stdlib.h>


#define keyport PORTD       //Keypad Port
#define keyportddr DDRD         //Data Direction Register
#define keyportpin PIND         //Keypad Port Pins

#define col1 PD0               //Column1 PortA.0
#define col2 PD1                //Column2 PortA.1
#define col3 PD2               //Column3 PortA.2
#define col4 PD3                //Column4 PortA.3

#define row1 PD4               //Column1 PortA.0
#define row2 PD5                //Column2 PortA.1
#define row3 PD6               //Column3 PortA.2
#define row4 PD7                //Column4 PortA.3

#define TRUE 1
#define FALSE 0

#define LcdDATA_DDR		DDRA			// Portul pe care conectam firele de date la LCD-ul
#define LcdDATA_PORT	PORTA
#define LcdDATA_PIN		PINA

#define LcdCMD_DDR		DDRA			// Portul pe care conectam firele de comenzi la LCD
#define LcdCMD_PORT		PORTA
#define LcdCMD_PIN		PINA

#define LcdD4					PA0				// Pin-ul pentru firul de date D4 de pe LCD
#define LcdD5					PA1				// Pin-ul pentru firul de date D5 de pe LCD
#define LcdD6					PA2				// Pin-ul pentru firul de date D6 de pe LCD
#define LcdD7					PA3				// Pin-ul pentru firul de date D7 de pe LCD

#define LcdRS					PA4			    // Pinul pentru selectare operatie (LCD)
#define LcdRW					PA5				// Pinul pentru Read/ Write (LCD)
#define LcdE					PA6				// Pinul de Enable (LCD)

#define RS_HIGH()				LcdCMD_PORT |= _BV(LcdRS)
#define RS_LOW()				LcdCMD_PORT &= ~_BV(LcdRS)
 
#define RW_HIGH()				LcdCMD_PORT |= _BV(LcdRW)
#define RW_LOW()				LcdCMD_PORT &= ~_BV(LcdRW)
 
#define ENABLE()				LcdCMD_PORT |= _BV(LcdE)
#define DISABLE()				LcdCMD_PORT &= ~_BV(LcdE)

#define LCD_INSTR_4wire 		0x28 	// 4 fire date, font 5x8
#define LCD_INSTR_display 		0x0C 	// Display On, Cursor On, Blinking On ( 1 Display Cursor Blink )
#define LCD_INSTR_clearDisplay  	0x01 	// Clear Display
#define LCD_INSTR_returnHome		0x02 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_nextLine 		0xC0 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_gotoCGRAM		0x40	// go to Character Generator RAM

void LCD_init();                // Initializare modul LCD.Trebuie apelata inainte de a se face orice operatie cu LCD-ul
void LCD_writeInstruction(unsigned char _instruction);	// Trimite o instructiune catre lcd (vezi datasheet)
void LCD_writeData(unsigned char _data);		// Trimite date catre LCD pentru afisare
void LCD_write(unsigned char _byte);			// trimite un bute catre LCD la modul general (nu conteaza daca e instructiune sau date)
void LCD_waitNotBusy();					// Asteptam pana cand lcd-ul devine disponibil pt o noua comanda
void LCD_waitInstructions(unsigned char _instructions);	// Asteapta un numar de cicli de ceas.
void LCD_print(char* _msg);				// Afiseaza imformatia pe LCD (doar 1 linie, primele 16 caractere)
void LCD_print2(char* _msg1, char* _msg2);		// Afisare pe 2 lini pe LCD
void LCD_printDecimal2u(unsigned int _n);		// Afisare numar in baza 10 pe LCD
void LCD_printHexa(unsigned int _n);			// Afisare numar in baza 16 pe LCD


/*****************************
 * IMPLEMENTARE API LCD.     *
 *****************************/

void LCD_init(){
	LcdDATA_DDR |=  (1<<LcdD4)|(1<<LcdD5)|(1<<LcdD6)|(1<<LcdD7);	// setam pinii de date ca pini de iesire
	LcdCMD_DDR  |=  (1<<LcdRS)|(1<<LcdRW)|(1<<LcdE);				// setam pinii de comenzi ca pini de iesire

	LCD_waitNotBusy();
	
	RS_LOW();												// Setam linia RS pe low
	RW_LOW();												// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD4)&~(1<<LcdD6)&~(1<<LcdD7);	// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |=  (1<<LcdD5);							// setam pinii de comenzi ca pini de iesire
	ENABLE();												// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	LCD_waitInstructions(6);								// Asteptam o perioada de timp T
	DISABLE();												// Setam linia E(nable) pe low; transferul s-a terminat
	LCD_writeInstruction(LCD_INSTR_4wire); 					// Incarcam comanda: 4 bit data, 2 lines, 5x8 font
	LCD_writeInstruction(LCD_INSTR_display); 				// Display On, Cursor On, Blinking On
	LCD_writeInstruction(0x06);								// Increment, no shift
	LCD_writeInstruction(0x01);								// Clear Display

	
}

void LCD_writeInstruction(unsigned char _instruction){
	/* TODO */					
	// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	// setam pinul RS pe low (low=instructiuni, high=date)
	// setam pinul RW pe low (suntem in modul de comenzi acum)
	// apelam procedura ce trimite byte-ul pe firele de date
	
	LCD_waitNotBusy();										// asteptam ca LCD-ul sa fie liber sa primeasca comenzile
	RS_LOW();												// setam pinul RS pe low.. low=instructiuni, high=date
	RW_LOW();												// setam pinu RW pe low (suntem in modul de comenzi acum)
	LCD_write(_instruction);
}


void LCD_writeData(unsigned char _data){
    /* TODO */
	// similar cu LCD_writeInstruction
	LCD_waitNotBusy();										// asteptam ca LCD-ul sa fie liber sa primeasca comenzile
	RS_HIGH();												// setam pinul RS pe high
	LcdCMD_PORT &= ~(1<<LcdRW);							// setam pinu RW pe low (suntem in modul de date acum)
	LCD_write(_data);

}

void LCD_write(unsigned char _byte){
    /* TODO */
	// scriem cei mai semnificativi 4 biti pe firele de date

	// Setam Pinul E pe high
	// Asteptam o perioada de timp T
	// Setam Pinul E pe low
	// Asteptam o perioada de timp T
	
	// scriem cei mai putin semnificativi 4 biti pe firele de date
	
	// Setam Pinul E pe high
	// Asteptam o perioada de timp T
	// Setam Pinul E pe low
	
	unsigned char MSB_byte;
	unsigned char LSB_byte;
	
	
	//scriem cei mai semnificativi 4 biti
	MSB_byte = (_byte>>4) & 0x0F;
	LcdDATA_PORT &= ~(1<<LcdD4); 							//setare la 0
	if (bit_is_set(MSB_byte, 0))
		LcdDATA_PORT |= (1<<LcdD4);						//setare la 1 daca bit-ul din byte-ul de date este setat
	LcdDATA_PORT &= ~(1<<LcdD5);
	if (bit_is_set( MSB_byte, 1))
		LcdDATA_PORT |= (1<<LcdD5);
	LcdDATA_PORT &= ~(1<<LcdD6);
	if (bit_is_set( MSB_byte, 2))
		LcdDATA_PORT |= (1<<LcdD6);
	LcdDATA_PORT &= ~(1<<LcdD7);
	if (bit_is_set(MSB_byte, 3))
		LcdDATA_PORT |= (1<<LcdD7);

	ENABLE();												// Setam Pinul E pe high
	LCD_waitInstructions(6);								// Asteptam o perioada de timp T
	DISABLE();												// Setam Pinul E pe low
	LCD_waitInstructions(6);								// Asteptam o perioada de timp T
	
	
	LSB_byte = _byte & 0x0F;
	LcdDATA_PORT &= ~(1<<LcdD4);
	if (bit_is_set(LSB_byte, 0))
		LcdDATA_PORT |= (1<<LcdD4);
	LcdDATA_PORT &= ~(1<<LcdD5);
	if (bit_is_set(LSB_byte, 1))
		LcdDATA_PORT |= (1<<LcdD5);
	LcdDATA_PORT &= ~(1<<LcdD6);
	if (bit_is_set(LSB_byte, 2))
		LcdDATA_PORT |= (1<<LcdD6);
	LcdDATA_PORT &= ~(1<<LcdD7);
	if (bit_is_set(LSB_byte, 3))
		LcdDATA_PORT |= (1<<LcdD7);

	ENABLE();												// Setam Pinul E pe high
	LCD_waitInstructions(6);								// Asteptam o perioada de timp T
	DISABLE();												// Setam Pinul E pe low
}

void LCD_waitNotBusy(){
	unsigned char _loop = 1;

	while (_loop)
	{
		LcdDATA_DDR &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7);		// Setam pinii de date de la LCD pe in pt a citi busy flag
		LcdDATA_PORT &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); 	// Dezactivam pullup resistor pentru pinii de in
		
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam pin-ul e pe low; ar trebui sa fie deja low, doar ne asiguram
		LcdCMD_PORT &= ~(1<<LcdRS);					// Setam pinul RS pe low
		LcdCMD_PORT |=  (1<<LcdRW);					// Setam pinul RW pe high (acum suntem in modul de interogare busy/adr)
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_loop = LcdDATA_PIN & (1<<LcdD7);				// Citim busy flag-ul
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low
		
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

		LcdDATA_DDR |= (1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); // Setam Portul de LCD ca port de iesire la loc
	}
}

void LCD_waitInstructions(unsigned char _instructions){
	while (_instructions--)
		;
}


void LCD_printDecimal2u(unsigned int _n){
	unsigned char tmp=0;
	
	// Extragem sutele
	while(_n>=100)
		_n-=100;

	while(_n>=10){
		tmp++;
		_n-=10;
	}

	LCD_writeData(tmp+'0');
	LCD_writeData(_n+'0');
}

void LCD_printHexa(unsigned int _n){
	unsigned char _tmp = _n>>4;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
	_tmp = _n & 0x0F;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
}


void LCD_print(char* _msg){
	unsigned char i=0;
	for( ; _msg[i]!=0 && i<16; i++)
		LCD_writeData( _msg[i] );
}

void LCD_print2(char* _msg1, char* _msg2){
	LCD_writeInstruction(LCD_INSTR_clearDisplay);
	LCD_print(_msg1);
	LCD_writeInstruction(LCD_INSTR_nextLine);
	LCD_print(_msg2);
}

unsigned char keyval;
void key_init(){
        keyportddr = 0xF0;
        keyport = 0x0F;
}


unsigned char get_key(){
        unsigned char i,key=1;
        for(i=0;i<4;i++){               //Loop for 4 rows
                keyport &=~(0x80>>i);   //Make rows low one by one
				_delay_ms(1);
                        if(!(keyportpin & (1<<col1))){
                         //check COL1
                                while(!(keyportpin & (1<<col1)));
                                //wait for release
                                return key;
                                //return pressed key value
                        }
                        if(!(keyportpin & (1<<col2))){
                         //Check COL2
                                key += 1;
                                //Second key pressed
                                while(!(keyportpin & (1<<col2)));
                                //wait for release
                                return key;
                                //return key value
                        }
                        if(!(keyportpin & (1<<col3))){
                         //Check COL3
                                key += 2;
                                //Third key pressed
                                while(!(keyportpin & (1<<col3)));
                                //Wait for release
                                return key;
                                //Return value
                        }
                        if(!(keyportpin & (1<<col4))){
                         //check COL4
                                key += 3;
                                //Fourth key pressed
                                while(!(keyportpin & (1<<col4)));
                                //Wait for release
                                return key;
                                //Return key value
                        }
                key +=4;        //Next row
                keyport |= 0x80>>i;
                //make read row high again
        }
        return FALSE;   //return false if no key pressed
}

unsigned char get_val_key(unsigned char key){
if(key==1) return (char)71;
else if(key==2) return '0';
else if(key==3) return (char)70;
else if(key==4) return '+';
else if(key==5) return '1';
else if(key==6) return '2';
else if(key==7) return '3';
else if(key==8) return '-';
else if(key==9) return '4';
else if(key==10) return '5';
else if(key==11) return '6';
else if(key==12) return '*';
else if(key==13) return '7';
else if(key==14) return '8';
else if(key==15) return '9';
else if(key==16) return '/';
return '=';
}

float rez_op (float nr2,float nr1,unsigned char op)
{
			if(op==43) 
			{
			
				nr2=nr1+nr2;	
			}
			if(op==45) 
			{
				nr2=nr1-nr2;
			}
			if(op==42) 
			{
				nr2=nr1*nr2;

			}
			if(op==47) 
			{
				nr2=nr1/nr2;
			}
return nr2;
}

int main()
{
float nr1=0;
float nr2=0;
char buf[10];
unsigned char op=20; 
int semn_ant=0;
int egal_ant=0;
int depasire=0;
LCD_init();
LCD_writeInstruction(LCD_INSTR_clearDisplay);
LCD_print("0");
key_init();

while(1){
keyval=get_key();
keyval=get_val_key(keyval);

if(depasire==0)
{
if((keyval<58)&&(keyval>47))                            //apasare tasta numerica
	{if(egal_ant==1)
		nr1=0;
	nr1=nr1*10+(keyval-48);
	LCD_writeInstruction(LCD_INSTR_clearDisplay);
	itoa(nr1,buf,10);
	LCD_print(buf);
	if(nr1>32000)
	{	
		LCD_writeInstruction(LCD_INSTR_clearDisplay);
		LCD_print("Overflow");
		depasire=1;
	}
	semn_ant=0;
	egal_ant=0;
	}

if((keyval==43)||(keyval==45)||(keyval==42)||(keyval==47))        //apasare +,-,*,/
	{
	if(semn_ant==0)
	{
	  if(op!=20)
		{
			nr2=rez_op (nr1,nr2,op);
			LCD_writeInstruction(LCD_INSTR_clearDisplay);
			
			itoa(nr2,buf,10);
			LCD_print(buf);
		if(nr2>32000)
	{	
		LCD_writeInstruction(LCD_INSTR_clearDisplay);
		LCD_print("Overflow");
		depasire=1;
	}
		}
		
		else
		{
		nr2=nr1;
		}
		nr1=0;
		semn_ant=1;
		egal_ant=0;
	}
		op=keyval;
	}

if(keyval==70)                               //apasare =
	{
	if(egal_ant==0)
	{
		nr2=rez_op (nr1,nr2,op);
		LCD_writeInstruction(LCD_INSTR_clearDisplay);
			itoa(nr2,buf,10);
			LCD_print(buf);
		if(nr2>32000)
	{	
		LCD_writeInstruction(LCD_INSTR_clearDisplay);
		LCD_print("Overflow");
		depasire=1;
	}
		nr1=nr2;
		op='0';
		semn_ant=0;
		egal_ant=1;
	}
	}
}	
if(keyval==71)                //apasare tasta C
	{
		LCD_print(buf);
		nr1=0;
		nr2=0;
		op='0';
		LCD_writeInstruction(LCD_INSTR_clearDisplay);
		itoa(nr2,buf,10);
		LCD_print(buf);
		semn_ant=0;
		egal_ant=0;
		depasire=0;
	}
	

	keyval=0;
}
return 0;
}